(************************************************************ 
 Io Environment Handling                                      
   Amalthea, io_messages.ml                                   
   Martin Sandin (d97masa@dtek.chalmers.se)                   
 ************************************************************)
open Io_types;;
open Io_parser;;


let variable_flag = ref false

let exported_ids = function 
  | (_, [], ds, _) -> List.map (fun (_, i, _) -> i) ds
  | (_, es, _, _)  -> es

(* The standard environment *)
let standard_environment : io_env
= ([], [])

(* Environment handling functions *)
let lookup = fun (en, ven) i ->
  !(List.assoc i en)

let vlookup = fun (_, ven) vi ->
  !(List.assoc vi ven)
  
let safe_lookup = fun en i ->
  try 
    lookup en i
  with Not_found ->
    raise (Io_interpreter_error (Io_undefined_error i))

let bind = fun (en, ven) (i, c) -> ((i, ref c)::List.remove_assoc i en, ven)
let vbind = fun (en, ven) (vi, c) -> (en, (vi, ref c)::List.remove_assoc vi ven)

let update_ref = fun (en, _) (i,c) ->
  (List.assoc i en) := c

let update_vref = fun (_, ven) (i,c) ->
  (List.assoc i ven) := c

(* Builds a continuation from an expression and a environment.
   decl: true for declared continuations, false for runtime constructed.
         This is used by the trace printer. *)
let build_cont
  : bool -> io_env -> io_expr -> io_cont
= fun decl en ex -> match ex with
  | Eint (p,n) -> Cint n
  | Estring (p,s) -> Cstring s
  | Eid (p,i) -> safe_lookup en i
  | Eprimitive pi -> Cprim (List.assoc pi Io_primitives.map)
  | _  -> Cexpr (ex, en, decl)


(* Finds free variables and returns the expressin with an env filter. *)
let rec massage_expr = function
  | Elambda (p, is, ex, _) ->
    let (fis,e) = massage_expr ex in
    let fis = List.filter (fun i -> not (List.mem i is)) fis in
    (fis,Elambda (p, is, e, EFlimit fis))
  | Eappl (p, i, es, _) ->
    let (iss, es) = List.split (List.map massage_expr es) in
    let iss = iss in
    let fis =
      List.fold_left (fun ais is ->
        ais @ (List.filter (fun i ->
          not (List.mem i ais)) is)) [] iss in
    (i::fis,Eappl (p, i, es, EFlimit fis))
  | Egetvar (p, vi, i, ex, _) ->
    if !variable_flag then
      let (fis, e) = massage_expr ex in
      let fis = List.filter (fun li -> li != i) fis in
      (fis, Egetvar (p, vi, i, e, EFlimit fis))
    else
      massage_expr ex
  | Eputvar (p, vi, a, ex, _) ->
    if !variable_flag then
      let (fis, e) = massage_expr ex in
      let (afis, ae) = massage_expr a in
      let fis = List.fold_left (fun is i -> if List.mem i is then is else i::is) fis afis in
      (fis, Eputvar (p, vi, ae, e, EFlimit fis))
    else
      massage_expr ex
  | (Eid (p, i)) as ex -> ([i], ex)
  | ex -> ([], ex)


(* Build local continuation and references it in the environment *)
let build_local_conts : io_env -> unit
= fun ((en, ven) as ten) ->
  let build (i, c) =
    match !c with
      | Clocal ex -> update_ref ten (i, build_cont true ten (snd (massage_expr ex)))
      | _ -> ()
  in let vbuild (i, c) =
    match !(!c) with
      | Clocal ex -> update_vref ten (i, ref (build_cont true ten (snd (massage_expr ex))))
      | _ -> ()
  in
    List.iter build en;
    List.iter vbuild ven


(* References an external continuation i in en, hunts it down in the list
   of environments, ens. *)
let link_external_conts : (string * io_env) list -> io_env -> unit
= fun ens ((en, _) as ten)->
  let link (i, c) = 
    let rec get_cont visited_ms c = match c with
      | Cexternal mname ->
          (try
            if List.mem mname visited_ms then
              raise Not_found
            else
              get_cont (mname :: visited_ms) (lookup (List.assoc mname ens) i)
          with
            Not_found ->
              raise (Io_module_error (Io_undefined_export_error (i, mname))))
      | c -> c
    in match !c with
      | Cexternal mname -> update_ref ten (i, (get_cont [] (lookup ten i)))
      | _ -> ()
  in
  List.iter link en
