(************************************************************ 
 Io Error and Trace Messages                                  
   Amalthea, io_messages.ml                                   
   Martin Sandin (d97masa@dtek.chalmers.se)                   
 ************************************************************)

open Io_types;;

(* Util *)
let string_of_pos : pos -> string
= function (m,r,c) ->
  Filename.basename m ^ " r" ^ string_of_int r ^ " c" ^ string_of_int  c

(* Returns the size of a continuation as a tuple of the number of
   continuations bound in recursive environments and the number of
   conts bound in this continuations environment. Declared and
   primitive conts excluded. *)
let rec cont_size = function
  | Cexpr (_,(en, ven),false) ->
    (1 + (List.fold_left (fun sum (_,c) -> sum + (fst (cont_size !c))) 0 en), List.length en)
  | Cexpr (_,_,true) -> (0,0)
  | _ -> (1,0)


(* Trace messages 
 ************************************************************)

(* Prints continuations to the desired depth *)
let rec print_cont
= fun level c -> match c with
  | Cexpr (ex, _, _) -> print_expr level ex
  | Cint n -> string_of_int n
  | Cstring s -> "\"" ^ s ^ "\""
  | Cprim _ -> "<primitive>"
  | _ -> raise (Io_error "Trace error: erronous continuation.")
  
and print_expr : int -> io_expr -> string
= fun level e -> match level with
  | 0 -> "..."
  | level -> match e with
    | Elambda (_,is,e,_) -> "-> "
        ^ String.concat " " is
        ^ ";" ^ print_expr (level - 1) e
    | Eappl (_,i,es,_) -> i ^ print_applparams (level - 1) es
    | Eputvar (_,i,e,e2,_) -> i ^ " <<=" ^ print_param (level - 1) e ^ ";" ^ print_expr (level - 1) e2
    | Egetvar (_,i,i2,e2,_) -> i ^ " =>> " ^ i2 ^ ";" ^ print_expr (level - 1) e2
    | Eid (_,i) -> i
    | Eint (_, n) -> string_of_int n
    | Estring (_,s) -> "\"" ^ s ^ "\""
    | Eprimitive i -> raise Amalathea_AST_error

and print_applparams: int -> io_expr list -> string
= fun level ps -> match ps with
    | [] -> ""
    | ((Elambda _) as e)::[] -> " " ^ print_expr level e
    | e::[] -> ";" ^ print_expr level e
    | e::es -> print_param level e ^ print_applparams level es

and print_param: int -> io_expr -> string
= fun level ps -> match ps with
    | (Eid (_,i)) -> " " ^ i
    | (Eint (_,n)) -> " " ^ string_of_int n
    | (Estring (_,s)) -> " \"" ^ s ^ "\""
    | e -> " (" ^ print_expr level e ^ ")"

(* Prints lambda expression trace messages *)
let print_lambda_trace
= fun p cmb ->
  print_endline ("  " ^ string_of_pos p ^ ": lambda called");
  List.iter (function (i,c) ->
    let (s,es) = cont_size c in
    print_endline ("    " ^ i
      ^ " := " ^ print_cont 3 c
      ^ " [" ^ string_of_int (es) ^ "/" ^ string_of_int (s) ^ "]")
  ) cmb

let print_enfilter_trace
= fun fltis -> print_endline ("    {" ^ String.concat ", " fltis ^ "}")

let print_application_trace 
= fun pos i -> print_endline ("  " ^ string_of_pos pos ^ ": application of '" ^ i ^ "'")

let print_primitive_trace 
= fun pos -> print_endline ("  " ^ string_of_pos pos ^ ": primitive call")


(* Module error messages 
 ************************************************************)
let parser_error
= fun p -> "Parse error after " ^ string_of_pos p

let lexer_error
= fun p -> "Parse error: invalid character after " ^ string_of_pos p

let module_not_found_error
= fun importer imported ->
  "Module error: "
  ^ "could not locate " ^ imported
  ^ " imported by " ^ importer
  
let undefined_export_error
= fun exported mname ->
  "Module error: "
  ^ "exported identifier undefined in "^Filename.basename mname
  ^": "^exported
  
let undefined_import_error
  = fun imported mname ->
    "Module error: "
    ^ "imported identifier undefined in "^Filename.basename mname
  ^": "^imported
  
  
(* Interpreter error messages 
 ************************************************************)
let parameter_mismatch_error
= fun apl_pos (form, lambda_p, (exp, got)) ->
  "Io error: call at " ^ string_of_pos apl_pos
  ^ " to " ^ form ^ " at " ^ string_of_pos lambda_p ^ ": "
  ^ "expected " ^ string_of_int exp
  ^ " parameter(s), got " ^ string_of_int got
  
let id_undefined_error
= fun pos i ->
  "Io error: " ^ i ^ " undefined in appl at "
  ^ string_of_pos pos
  
let primitive_error
= fun m ->
  "Io primitive error: " ^ m